from typing import Optional, List, cast

import pygame
from pygame.constants import HWSURFACE, DOUBLEBUF, RESIZABLE
from pygame.surface import Surface

from tools.vector import vectorDivF, vectorMulI
from .IUIEventHandler import IUIEventHandler
from .Mouse import Mouse
from .MouseWheel import MouseWheel
from .component.Component import Component
from .component.text import TextRenderer
from .mode import EditGameMode
from .mode.GameMode import GameMode
from .theme.Theme import Theme


class UserInterface:
    def __init__(self, theme: Theme):

        # Create window with default resolution
        pygame.init()
        windowSize = vectorMulI(theme.viewSize, (3, 3))
        self.__window = pygame.display.set_mode(windowSize, HWSURFACE | DOUBLEBUF | RESIZABLE)
        pygame.display.set_caption("2D Medieval Strategy Game with Python, http://www.patternsgameprog.com")
        pygame.display.set_icon(pygame.image.load("assets/toen/icon.png"))

        # Rendering
        self.__theme = theme
        self.__rescaledShift = (0, 0)
        self.__rescaledScale = (1.0, 1.0)
        self.__font = theme.getFont("default")

        # Inputs
        self.__mouseFocus: Optional[Component] = None

        # Other
        self.__gameMode: Optional[GameMode] = None
        self.__running = True
        self.__clock = pygame.time.Clock()

    @property
    def theme(self) -> Theme:
        return self.__theme

    def setGameMode(self, gameMode: GameMode):
        self.__gameMode = gameMode

    def __processKeyEvent(self, event):
        if event.type == pygame.KEYDOWN:
            self.handleKeyDown(event.key)
        elif event.type == pygame.KEYUP:
            self.handleKeyUp(event.key)

    def __createMouse(self) -> Mouse:
        mouseX, mouseY = pygame.mouse.get_pos()
        mouseX = int((mouseX - self.__rescaledShift[0]) / self.__rescaledScale[0])
        mouseY = int((mouseY - self.__rescaledShift[1]) / self.__rescaledScale[1])
        pygameButtons = pygame.mouse.get_pressed(num_buttons=3)
        return Mouse((mouseX, mouseY), pygameButtons)

    def __updateMouseFocus(self):
        # Mouse
        if pygame.mouse.get_focused():
            mouse = self.__createMouse()
            newFocus = self.findMouseFocus(mouse)
        else:
            mouse = Mouse((-10000, -10000))
            newFocus = None
        if newFocus != self.__mouseFocus:
            if self.__mouseFocus is not None \
                    and isinstance(self.__mouseFocus, IUIEventHandler):
                self.__mouseFocus.mouseLeave()
            self.__mouseFocus = newFocus
            if self.__mouseFocus is not None \
                    and isinstance(self.__mouseFocus, IUIEventHandler):
                self.__mouseFocus.mouseEnter(mouse)

    def __processMouseEvent(self, event):
        mouse = self.__createMouse()
        viewSize = self.__theme.viewSize
        if 0 <= mouse.pixel[0] < viewSize[0] \
                and 0 <= mouse.pixel[1] < viewSize[1]:
            if event.type == pygame.MOUSEBUTTONDOWN:
                self.handleMouseButtonDown(mouse)
            elif event.type == pygame.MOUSEBUTTONUP:
                self.handleMouseButtonUp(mouse)
            elif event.type == pygame.MOUSEWHEEL:
                wheel = MouseWheel(event.x, event.y, event.flipped, event.which)
                self.handleMouseWheel(mouse, wheel)
            elif event.type == pygame.MOUSEMOTION:
                self.handleMouseMove(mouse)

    def processInput(self):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.__running = False
                break
            elif event.type == pygame.KEYDOWN \
                    or event.type == pygame.KEYUP:
                if event.key == pygame.K_ESCAPE:
                    self.__running = False
                    break
                self.__processKeyEvent(event)
            elif event.type == pygame.MOUSEBUTTONDOWN \
                    or event.type == pygame.MOUSEBUTTONUP \
                    or event.type == pygame.MOUSEWHEEL \
                    or event.type == pygame.MOUSEMOTION:
                self.__processMouseEvent(event)
        self.__updateMouseFocus()
        if self.__gameMode is not None:
            self.__gameMode.processInput()

    def update(self):
        if self.__gameMode is not None:
            self.__gameMode.update()

    def render(self):
        # Render world in a surface
        renderSurface = Surface(self.__theme.viewSize)
        if self.__gameMode is not None:
            self.__gameMode.render(renderSurface)

        # Scale rendering to window size
        windowWidth, windowHeight = self.__window.get_size()
        viewSize = self.__theme.viewSize
        renderRatio = viewSize[0] / viewSize[1]
        windowRatio = windowWidth / windowHeight
        if windowRatio <= renderRatio:
            rescaledSize = (windowWidth, int(windowWidth / renderRatio))
            self.__rescaledShift = (0, (windowHeight - rescaledSize[1]) // 2)
        else:
            rescaledSize = (int(windowHeight * renderRatio), windowHeight)
            self.__rescaledShift = ((windowWidth - rescaledSize[0]) // 2, 0)

        # Scale the rendering to the window/screen size
        rescaledSurface = pygame.transform.scale(renderSurface, rescaledSize)
        self.__rescaledScale = vectorDivF(rescaledSurface.get_size(), renderSurface.get_size())
        self.__window.blit(rescaledSurface, self.__rescaledShift)

        # Draw frame time
        frameTime = self.__clock.get_rawtime()
        textSurface = self.__font.render(f"{frameTime}ms", False, (255, 255, 255), (0, 0, 0))
        self.__window.blit(
            textSurface, (
                self.__window.get_width() - textSurface.get_width(),
                self.__window.get_height() - textSurface.get_height()
            )
        )

    def run(self):
        # Main game loop
        while self.__running:
            self.processInput()
            self.update()
            self.render()

            pygame.display.update()
            self.__clock.tick(30)

        if self.__gameMode is not None:
            self.__gameMode.dispose()

    # Quit

    def quit(self):
        pygame.quit()

    # Event handling

    def getEventHandlers(self) -> List[GameMode]:
        return [mode for mode in [
            self.__gameMode
        ] if mode is not None]

    def handleKeyDown(self, key: int) -> bool:
        for handler in self.getEventHandlers():
            if handler.keyDown(key):
                return True
        return False

    def handleKeyUp(self, key: int) -> bool:
        for handler in self.getEventHandlers():
            if handler.keyUp(key):
                return True
        return False

    def findMouseFocus(self, mouse: Mouse) -> Optional[Component]:
        for handler in self.getEventHandlers():
            focus = handler.findMouseFocus(mouse)
            if focus is not None:
                return focus
        return None

    def handleMouseButtonDown(self, mouse: Mouse) -> bool:
        for handler in self.getEventHandlers():
            if handler.mouseButtonDown(mouse):
                return True
        return False

    def handleMouseButtonUp(self, mouse: Mouse) -> bool:
        for handler in self.getEventHandlers():
            if handler.mouseButtonUp(mouse):
                return True
        return False

    def handleMouseWheel(self, mouse: Mouse, wheel: MouseWheel) -> bool:
        for handler in self.getEventHandlers():
            if handler.mouseWheel(mouse, wheel):
                return True
        return False

    def handleMouseMove(self, mouse: Mouse) -> bool:
        for handler in self.getEventHandlers():
            if handler.mouseMove(mouse):
                return True
        return False

