from typing import Tuple, Optional, List

from core.state import World, Layer, ILayerListener
from tools.vector import vectorDivI, vectorAddI
from .LayerComponent import LayerComponent
from .LayerComponentFactory import LayerComponentFactory
from ..CompositeComponent import CompositeComponent
from ...IUIEventHandler import IUIEventHandler
from ...Mouse import Mouse
from ...theme.Theme import Theme


class WorldComponent(CompositeComponent, IUIEventHandler, ILayerListener):

    def __init__(self, theme: Theme, world: World):
        super().__init__(theme)
        self.__world = world
        self.__view = (0, 0)
        self.__previousCell: Optional[Tuple[int, int]] = None
        self.__mouseButtonDown = False

        self.__layers: List[Layer] = []
        self.__layerComponents: List[LayerComponent] = []
        factory = LayerComponentFactory(theme, world)
        for name in world.layerNames:
            layerComponent = factory.create(name)
            self.addComponent(layerComponent, cache=True)
            self.__layerComponents.append(layerComponent)

            layer = world.getLayer(name)
            layer.registerListener(self)
            self.__layers.append(layer)

        self.__tileSize = theme.getTileset(world.layerNames[0]).tileSize

    def dispose(self):
        for layer in self.__layers:
            layer.removeListener(self)
        super().dispose()

    @property
    def view(self) -> Tuple[int, int]:
        return self.__view

    # Layer listener

    def cellChanged(self, layer: Layer, cell: Tuple[int, int]):
        for layerComponent in self.__layerComponents:
            layerComponent.cellChanged(layer, cell)

    # Component listener

    def viewChanged(self, view: Tuple[int, int]):
        super().viewChanged(view)
        self.__view = view

    # UI Event Handler

    def __computeCellCoordintates(self, pixel: Tuple[int, int]) -> Optional[Tuple[int, int]]:
        pixel = vectorAddI(pixel, self.__view)
        cell = vectorDivI(pixel, self.__tileSize)
        if not self.__world.contains(cell):
            return None
        return cell

    def mouseButtonDown(self, mouse: Mouse) -> bool:
        super().mouseButtonDown(mouse)
        cell = self.__computeCellCoordintates(mouse.pixel)
        if cell is None:
            return False
        self.__mouseButtonDown = True
        self.notifyWorldCellClicked(cell, mouse)
        return True

    def mouseMove(self, mouse: Mouse) -> bool:
        cell = self.__computeCellCoordintates(mouse.pixel)
        if cell is None:
            self.__previousCell = None
            return False
        if self.__previousCell is not None and cell == self.__previousCell:
            return False
        self.__previousCell = cell
        self.notifyWorldCellEntered(cell, mouse, self.__mouseButtonDown)
        return True

    def mouseButtonUp(self, mouse: Mouse) -> bool:
        self.__mouseButtonDown = False
        return True

    def mouseEnter(self, mouse: Mouse) -> bool:
        self.__mouseButtonDown = False
        return True

    def mouseLeave(self) -> bool:
        self.__mouseButtonDown = False
        return True
